// $Id: Connection.cc,v 1.31.2.4 1998/10/16 02:20:43 rousskov Exp $

/*
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation, and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "AppDef.h"

#include <sys/types.h>
#include <sys/uio.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>

#include "Connection.h"
#include "StatShooters.h"
#include "ErrHist.h"
#include "XactLog.h"
#include "Log.h"
#include "Gadgets.h"

extern EventShortCounter stOpenCount;
extern EventShortCounter stOpenErrCount;
extern EventCounter stErrCount;
extern LevelMeter stConnLevel;

bool Connection::open() {
	assert(theAddr.host() && theAddr.port());
	if (!theSocket.create()) {
		TheErrHist[errno]++;
		SysLog << "connection " << this << " failed to create a socket; cause(?): " << strerror(errno) << ende;
		stOpenErrCount++;
		return false;
	}

	theSocket.reuseAddr(true);
	theSocket.reusePort(true);
	theSocket.blocking(false);
	theSocket.linger(0);

	Journal << "connection " << this << " fd: " << (int)theSocket << " ~opened to " << theAddr << ende;
	// bind to an ephemeral port then connect
	const bool res = theSocket.bind(0) && theSocket.connect(theAddr.host(), theAddr.port());
	if (res) {
		stOpenCount++;
		stConnLevel++;
	} else {
		stOpenErrCount++;
		theSocket.close(); // close here
	}
	//theIOReadyFlags[ioWrite] = res;
	return res;
}

bool Connection::accept(Socket listenSock, bool &fatal) {
	struct sockaddr_in addr;
	int addr_len = sizeof(addr);
	theSocket = listenSock.accept((struct sockaddr*)&addr, &addr_len);
	if (theSocket >= 0) {
		theAddr.host(inet_ntoa(addr.sin_addr));
		theAddr.port(addr.sin_port);
		stOpenCount++;
		stConnLevel++;
		theIOReadyFlags[ioRead] = false; theIOReadyFlags[ioWrite] = true;
		Journal << "connection " << this << " fd: " << (int)theSocket << " accepted from " << theAddr << ende;
	} else {
		fatal = errno != EWOULDBLOCK;
		if (fatal) {
			TheErrHist[errno]++;
			stOpenErrCount++;
			SysLog << "connection " << this << " failed to accept from " << listenSock << " cause(?): " << strerror(errno) << ende;
		}
		Notepad << "connection " << this << " failed to accept from " << listenSock << " cause(?): " << strerror(errno) << ende;
	}
	return theSocket >= 0;
}

void Connection::close() {
	extern EventShortCounter stCloseCount;
	Journal << "connection " << this << " fd: " << (int)theSocket << " closed to " << theAddr << ende;
	theIOReadyFlags[ioRead] = theIOReadyFlags[ioWrite] = theIOReadyFlags[ioExcept] = false;
	theSocket.close();
	stCloseCount++;
	stConnLevel--;
}

// from the socket into the rBuf
Size Connection::read(Size max_size) {
	extern EventShortCounter stReadCount;

	theReaderBuf.stretch(max_size);
	assert(theIOReadyFlags[ioRead]); // remove if flags are not guaranteed to be set
	theIOReadyFlags[ioRead] = false;

	if (theRec) theRec->startRead();
	const ssize_t sz = ::read(theSocket, theReaderBuf.buf(), max_size);
	theReaderBuf.err(sz < 0 ? errno : 0);
	if (theRec) theRec->stopRead();

	Journal << "connection " << this << " fd: " << (int)theSocket << " read " << sz << " bytes (max: " << max_size << ")" << ende;
	theReaderBuf.eod(sz <= 0);
	if (sz > 0)
		theReaderBuf.appended(sz);
	if (sz < 0) {
		TheErrHist[errno]++;
		SysLog << "connection " << this << " read error: " << theReaderBuf.errNo() << ' ' << strerror(theReaderBuf.errNo()) << ende;
	}
	stReadCount++;

	// KLUDGE
	static int readCount = 10;
	if (readCount-- > 0) {
		struct sockaddr_in s;
		int len = sizeof(s);
		if (theSocket.peerAddr((struct sockaddr *) &s, len))
			cout << "# read[" << readCount << "] " << sz << " bytes from " << inet_ntoa(s.sin_addr) << ':' << s.sin_port << endl;
	}


	return sz;
}

// from the wBuf into the socket
Size Connection::write(Size max_size) {
	extern EventShortCounter stWriteCount;
	if (max_size <= 0)
		return 0;
	assert(max_size <= theWriterBuf.contSize());
	assert(theIOReadyFlags[ioWrite]); // remove if flags are not guaranteed to be set
	theIOReadyFlags[ioWrite] = false;

	if (theRec) theRec->startWrite();
	const ssize_t sz = ::write(theSocket, theWriterBuf.content(), max_size);
	theWriterBuf.err(sz < 0 ? errno : 0);
	if (theRec) theRec->stopWrite();

	Journal << "connection " << this << " fd: " << (int)theSocket << " wrote " << sz << " bytes (max: " << max_size << ")" << ende;
	if (sz > 0)
		theWriterBuf.consumed(sz);
	if (sz < 0) {
		TheErrHist[errno]++;
		SysLog << "connection " << this << " write error: " << theWriterBuf.errNo() << ' ' << strerror(theWriterBuf.errNo()) << ende;
	}
	stWriteCount++;
	return sz;
}
