// $Id: XactLog.cc,v 1.10.2.1 1998/09/22 07:43:46 rousskov Exp $

/*
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation, and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "XactLog.h"
#include "Gadgets.h"

#include <iomanip.h>

XactLog TheXactLog;

static const char *TheXactLogMagic = "XactLog";

/* XactRec::IOs */

void XactRec::IOs::save(ostream &os) const {
	os.write(&theCount, 1);
	for (int i = 0; i < theCount && i <= EXP_XIO_COUNT; i++) {
		shton32(os, theIOs[i].theStart);
		shton32(os, theIOs[i].theEnd);
	}
}

void XactRec::IOs::load(istream &is) {
	is.read(&theCount, 1);
	for (int i = 0; i < theCount && i <= EXP_XIO_COUNT; i++) {
		sntoh32(is, theIOs[i].theStart);
		sntoh32(is, theIOs[i].theEnd);
	}
}

void XactRec::IOs::print(ostream &os, int idx) const {
	if (idx < theCount && idx <= EXP_XIO_COUNT)
		os
			<< setw(3) << theIOs[idx].theStart
			<< "\t " 
			<< setw(3) << theIOs[idx].theEnd;
	else
		os
			<< setw(3) << -1
			<< "\t " 
			<< setw(3) << -1;
}

/* XactRec */

void XactRec::save(ostream &os) const {
	shton32(os, theId);
	shton32(os, theParentId);
	shtonTime(os, theStart);
	theReads.save(os);
	theWrites.save(os);
}

void XactRec::load(istream &is) {
	sntoh32(is, theId);
	sntoh32(is, theParentId);
	sntohTime(is, theStart);
	theReads.load(is);
	theWrites.load(is);
}

void XactRec::print(ostream &os, Time globalStart) const {

	const int lastIO = max(theReads.count(), theWrites.count()) - 1;
	const int maxIter = min(EXP_XIO_COUNT, lastIO);

	for (int i = 0; i <= maxIter; ++i) {
		os
			<< theId
			<< " \t" << theParentId
			<< " \t" << setw(10) << (theStart - globalStart).usec()
			<< " \t" << (i == maxIter ? lastIO : i);

		theReads.print(os << " \t", i);
		theWrites.print(os << " \t", i);
		os << endl;
	}
}

/* XactLog */

XactLog::XactLog(): 
	theStartCount(0), theSize(-1), theDebt(0), theCount(0), theLastId(0) {
}

void XactLog::configure(int aStart, int aSize) {
	Assert(!theCount);    // have not been active yet
	theStartCount = aStart+1;
	theSize = theDebt = aSize;
	stretch(theSize);
}

void XactLog::startNow() {
	Assert(theSize == theDebt); // not started or starting now
	Assert(theSize >= 0); // was configured
	theStartCount = theCount;
}

XactRec *XactLog::getRec() {

	theCount++;

	if (theDebt <= 0) // already done
		return 0;
	if (theStartCount <= 0) // will manually call startNow
		return 0;

	if (theStartCount > theCount) // not yet
		return 0;
	if (theStartCount == theCount) // now!
		startNow();

	// return next available record
	theDebt--;
	const int idx = Arr::theCount++;
	Assert(idx < theCapacity);
	XactRec &xr = item(idx);
	return &xr;
}

void XactLog::save(ostream &os) const {
	saveHeader(os);

	// records
	for (int i = 0; i < Arr::theCount; ++i)
		item(i).save(os);
}

void XactLog::load(istream &is) {
	const int count = loadHeader(is);
	if (count < 0)
		return;

	stretch(count);
	Arr::theCount = count;

	for (int i = 0; i < Arr::theCount && is.good(); ++i)
		item(i).load(is);
}

bool XactLog::print(istream &is, ostream &os) {
	const int count = loadHeader(is);
	if (count < 0)
		return false;

	printHeader(os, count);

	XactRec xr;
	if (count)
		xr.load(is);
	const Time globalStart = xr.start();
	os << "# start: " << globalStart << endl;

	// print table header
	os << "# xid"
		<< " \t" << "pid"
		<< " \t" << setw(10) << "dstart"
		<< " \t" << "iter"
		<< " \t" << "dRstrt"
		<< " \t" << "dRend"
		<< " \t" << "dWstrt"
		<< " \t" << "dWend"
		<< endl;

	for (int i = 0; i < count && is.good(); ++i) {
		if (i)
			xr.load(is);
		xr.print(os, globalStart);
	}

	return true;
}

void XactLog::saveHeader(ostream &os) const {
	// magic string
	os << TheXactLogMagic << endl;

	os
		<< theStartCount << endl
		<< theSize << endl
		<< theDebt << endl
		<< theCount << endl
		<< Arr::theCount << endl
		<< ends;
}

int XactLog::loadHeader(istream &is) {
	Assert(!theCount); // have not started yet

	// magic string
	if (!GetConstLine(is, TheXactLogMagic))
		return -1;

	int count;

	char c = '?';
	is 
		>> theStartCount >> ws
		>> theSize >> ws
		>> theDebt >> ws
		>> theCount >> ws
		>> count >> ws
		>> c;

	if (c != '\0')
		return -1; /* must end with ends */

	// success
	return count;
}

void XactLog::printHeader(ostream &os, int count) const {
	os << "# x-interval:"
		<< " start: " << theStartCount
		<< " end: " << theCount
		<< endl;
	os << "# size:"
		<< " wanted: " << theSize
		<< " debt: " << theDebt
		<< " got: " << count
		<< " volume: " << sizeof(XactRec) * theSize << " bytes"
		<< endl;
}

