// $Id: main.cc,v 1.49.2.9 1998/10/16 03:54:38 rousskov Exp $

/*
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation, and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "AppDef.h"

#include <stdlib.h>
#include <signal.h>
#include <new.h>

#include <fstream.h>
#include <iomanip.h>

#include "Size.h"
#include "Client.h"
#include "Server.h"
#include "AlarmClock.h"
#include "SelectLoop.h"
#include "ConnMgr.h"
#include "LoadGen.h"
#include "ConstAddrGen.h"
#include "LimitPgsMeter.h"
#include "ConstAddrRndReqMaker.h"
#include "RndRepMaker.h"
#include "RepEater.h"
#include "ReqEater.h"
#include "BlockReader.h"
#include "BlockWriter.h"
#include "LocalEvents.h"
#include "Array.h"
#include "Shooter.h"
#include "Stats.h"
#include "Trace.h"
#include "ErrHist.h"
#include "PhaseMgr.h"
#include "XactLog.h"
#include "Log.h"
#include "Gadgets.h"
#include "globals.h"


// local constants
static const Time TheMaxSelectDelay = Time(300,0); // 5 min

// local vars

// true if we should exit main loop nomatterwhat
static bool TheShutDownFlag = false;

// import
extern RateMeter TheSubmMeter;
extern RateMeter TheComplMeter;
extern RateMeter TheErrMeter;

// XXX: big kludge
Trace *TheRespTimeTrace = 0;
Trace *TheRespSizeTrace = 0;

// saves the values of real meters at the end of the measurement phase
static PhaseMgr ThePhaseMgr;


// main loop: iterate until no alarms are set and select loops is idle
static 
void loop() {
	//extern ValueStore stEvQueueDelay;
	//extern ValueStore stMainLoopDelay;
	extern EventCounter stEventCount;
	extern bool StatsIdle();

 	while (!TheAlarmClock.isClear() || !TheSelectLoop.isIdle() || !TheEventQueue.empty() || !TheKernel.idle()) {
		UpdateClocks();
		//const Time iterStart = TheClock;

		while (!TheEventQueue.empty()) {
			stEventCount += TheEventQueue.step();
			UpdateClocks();
		}
		//stEvQueueDelay = (TheClock - iterStart).usec();

		if (!TheKernel.idle()) {
			TheKernel.slice();
			if (!TheKernel.idle()) // pause: check alarms and events 
				continue;
		}

		UpdateClocks();

		// save stats for this iteration if there was an activity
		//if (!StatsIdle())
		//	TheShotRegistry->snapShot(TheClock);

		// if alarm is not set, use the default delay
		Time time_left = TheAlarmClock.isClear() ? 
			TheMaxSelectDelay : TheAlarmClock.timeLeft();

		// call select even if the SelectLoop is idle (but there are alarms)
		if (!TheSelectLoop.isIdle() || !TheAlarmClock.isClear()) {
			const int count = TheSelectLoop.step(time_left);
			// stop if busy select returned nothing after max delay
			if (!count && !TheSelectLoop.isIdle() && TheAlarmClock.isClear() && time_left >= TheMaxSelectDelay) {
				TheShutDownFlag = true;
				SysLog << "# shutting down after " << time_left << " of inactivity" << ende;
			}
		}

		ThePhaseMgr.checkPhase();

		// break if shut down requested
		if (TheShutDownFlag)
			break;
	}
}

static
void outOfMem() {
	cerr << "out of memory in new" << endl;
	abort();
}

static
void reportRespTimeAndSize() {

	if (!TheRespTimeTrace || !TheRespSizeTrace) {
		Report << "# response time and size stats disabled" << endl;
		return;
	}

	TheRespTimeTrace->report(Report << endl);
	TheRespSizeTrace->report(Report << endl);

	// report response time versus size
	Array<int> perc;
	TheRespSizeTrace->perc(perc); // get percentiles
	if (!perc.count()) {
		Report << "# response time and size stats not available" << endl;
		return;
	}
	Assert(perc.count() == 101);

	// print header
	Report << endl << "Resp Time for major Resp Size ranges: {" << endl
		<< setw(6) << "perc"
		<< "\t " << setw(6) << "min_sz"
		<< "\t " << setw(6) << "max_sz";
	for (int j = 0; j <= 100; j += 25)
		Report << "\t " << setw(5) << j << '%';
	Report << endl;

	// XXX: a very inefficient and ugly loop
	for (int p = 10; p <= 100; p += 10) {
		const int min_size = perc[p-10];
		const int max_size = perc[p] - (int)(p < 100);

		// collect matching response times
		const int kept_count = TheRespTimeTrace->keptCount();
		Assert(kept_count == TheRespSizeTrace->keptCount());
		Trace T("tmp", "usec", kept_count);
		for (int i = 0; i < kept_count; ++i) {
			const int sz = TheRespSizeTrace->item(i);
			if (sz >= min_size && sz <= max_size)
				T.record(TheRespTimeTrace->item(i));
		}

		// print percentiles for this size range
		if (T.stat().count()) {
			Array<int> tp;
			T.perc(tp);
			if (!tp.count())
				continue;
			Assert(tp.count() == 101);
			Report << setw(6) << p
				<< "\t " << setw(6) << min_size 
				<< "\t " << setw(6) << max_size;
			for (int j = 0; j <= 100; j += 25)
				Report << "\t " << setw(6) << tp[j];
			Report << endl;
		}
	}
	Report << '}' << endl;
}

static
void report(int id, const RateMeter &subm, const RateMeter &compl, const RateMeter &err) {
	if (id >= 0)
		Report << setw(6) << id;
	else
		Report << (id == -1 ? "middle" : "sum");

	Report
		<< "\t " << subm.count()
		<< "\t " << subm
		<< "\t " << compl.count()
		<< "\t " << compl
		<< "\t " << err.count()
		<< "\t " << (subm.count() - compl.count())
		<< endl;
}

template <class TopProc>
static
void report(const char *prefix, const Array<TopProc*> &procs) {
	Report << endl;
	if (!procs.count()) {
		Report << "# " << prefix << ": none" << endl;
		return;
	}

	// header
	Report << prefix
		<< "\t subm"
		<< "\t s/sec"
		<< "\t compl"
		<< "\t c/sec"
		<< "\t errs"
		<< "\t left"
		<< endl;
	// report for all processes
	RateMeter submM, complM, errM;
	for (int i = 0; i < procs.count(); ++i) {
		complM += procs[i]->complMeter();
		submM += procs[i]->submMeter();
		errM += procs[i]->errMeter();
		procs[i]->report(Report << setw(6) << i << "\t ");
		Report << endl;
	}
	// totals
	report(-2, submM, complM, errM);
}

static
void shutDown(int) {
	TheShutDownFlag = true;
	signal(SIGINT, SIG_DFL);
}

static
void configure() {

	// set resource limits
	TheMaxFD = SetMaxFD(TheMaxFD);
	assert(TheMaxFD > 0);

	// ignore stupid signals
	signal(SIGPIPE, SIG_IGN);

	// quit nicely on SIGINT
	signal(SIGINT, &shutDown);

	// custom out-of-memory handler
	set_new_handler(&outOfMem);

	// log format
	ConfigureLogs(2);

	// world params
	TheWorld.id(TheOpt.theWorldId);
	TheWorld.capacity(TheOpt.theWorldCapacity);
	TheWorld.objSizeMean(TheOpt.theObjSizeMean);
}



template <class DB>
static
void saveDB(const DB &db, const char *name, const char *fname) {
	Journal << "saving " << name << " in " << fname << ende;

	ofstream of(fname);
	db.save(of);
	if (!of.good())
		SysLog << "error writing " << name << " to " << fname << " cause(?): " << strerror(errno) << ende;
	of.close();
}

static
void saveResults() {
	// XXX: get a better fnaming scheme!
	saveDB(*TheShotRegistry, "stat snapshots", "/tmp/t.shot");
	saveDB(TheXactLog, "xact log", "/tmp/t.xlog");
}

static
int usage(const char *prg_name) {
	static char *Id = StrDup("Polygraph. Version 0.0p1. $Date: 1998/10/16 03:54:38 $ (c) 1998");

	for (char *p = Id; *p; ++p)
		if (*p != '$') cerr << *p;
	cerr << endl;

	cerr << "usage: " << prg_name
		<< " [option ...]"
		<< " <clients>"
		<< " [listen_port# ...]"
		<< endl;

	cerr << "clients:" << endl
		<< "  <num>x<count>         'stream' workload:  num clients, count requests each" << endl
		<< "                          requests are submitted as replies are received" << endl
		<< "  <num>x<count>r<rate>  'Poisson' workload: cumulative req rate" << endl
		<< "                          rate is an average submission rate in req/sec" << endl
		<< "  <0>                   no clients (server-only mode)" << endl;

	cerr << "options:" << endl
		<< "  --host <name>         origin server host (affects URL generation)" << endl
		<< "  --proxy <IP:port#>    proxy address (affects connect(2) requests)" << endl
		<< "  --world <id:count>    numerical world id and #objects in the world" << endl
		<< "                          use 'max' for maximum #objects" << endl
		<< "  --seq <seed>          sequential object ids; ids start with seed" << endl
		<< "  --rnd <seed>          random object ids; r.n.g seeded with seed" << endl
		<< "                          ids are distributed as zipf(1)" << endl
		<< "  --unique <seed>       append unique ids to urls; ids start with seed" << endl
		<< "  --obj_size <kbytes>   mean object size" << endl
		<< "  --serv_think <msec>   server think time (per write)" << endl;

	cerr << "URL format (abridged):" << endl
		<< "  http://host_name/obj_id/world_id" << endl
		<< " or" << endl
		<< "  http://host_name/obj_id/world_id/unique_id" << endl;

	return -1;
}

int main(int argc, char *argv[]) {

	if (!TheOpt.parse(argc, argv))
		return usage(argv[0]);

	configure();

	UpdateClocks();
	Time startTime = TheClock;

	Journal << ThePrgName << " started." << ende;
	Report << "Command:";
	for (int i = 0; i < argc; ++i)
		Report << ' ' << argv[i];
	Report << endl;

	TheOpt.report(Report);
	Report << "Max_FD: " << TheMaxFD << endl;
	
	// proxy address (affects client connect requests)
	const Address &addrProxy = TheOpt.theProxyAddr;

	// default server configuration
	ServerCfg scfg;
	scfg.theConnMgr = new ConnMgr(0);
	scfg.thePause = new Pause();
	scfg.theXactCfg.theContMaker = new RndRepMaker(Size::KB(4), false);
	scfg.theXactCfg.theReader = new BlockReader(Size::KB(4));
	scfg.theXactCfg.theWriter = new BlockWriter(Size::KB(4));
	scfg.theXactCfg.theContEater = new ReqEater();

	const int client_count = TheOpt.theClientCount;
	const int server_count = TheOpt.theServerCount;
	const IntArray &ports = TheOpt.theServerPorts;

	// configuration for the client's connection mgr
	ConnMgrCfg cConnMgrCfg;
	cConnMgrCfg.theMaxCcUseLevel = 1; /* no cc requests on one connection */
	cConnMgrCfg.theMaxTotUseCount = 1; /* no persistent connections at all */

	// progress meter configuration for clients
	const int req_per_client = TheOpt.theReqPerClient;
	ProgMeterCfg pmCfg;
	pmCfg.theMaxCcLevel = 16; // max cc requests at a time
	pmCfg.theMaxStartCount = req_per_client;

	ThePhaseMgr.goal(req_per_client * client_count);

	// loader configuration for clients
	Loader *loader = 0;
	if (TheOpt.theClientThinkTime >= 0) { // LoadGen ("real")
		LoadGenCfg cfgLoadGen;
		cfgLoadGen.theBlobSize = 1;
		cfgLoadGen.theBlobCount = -1; /* no limit */
		cfgLoadGen.theIntraBlobPause = new ConstPause(Time::Sec(0));
		cfgLoadGen.theInterBlobPause =
			new RndPause(new ExpDistr(new RndGen(), TheOpt.theClientThinkTime));
		LoadGen *lg = new LoadGen();
		lg->configure(cfgLoadGen);
		loader = lg;
	} else {
		loader = new StreamLoader(1);
	}
	assert(loader);

	// default client configuration
	ClientCfg ccfg;
	ccfg.theConnMgr = new ConnMgr(0);
	ccfg.theConnMgr->configure(cConnMgrCfg);
	ccfg.theProgressMeter = new ProgressMeter();
	ccfg.theProgressMeter->configure(pmCfg);
	ccfg.theLoader = loader;
	ccfg.theAddrGen = new ConstAddrGen(addrProxy);
	ccfg.theXactCfg.theContMaker = new ConstAddrRndReqMaker(TheOpt.theOriginAddr);
	ccfg.theXactCfg.theReader = new BlockReader(Size::KB(4));
	ccfg.theXactCfg.theWriter = new BlockWriter(Size::KB(4));
	ccfg.theXactCfg.theContEater = new RepEater();

	// additional configuration with req_per_client known
	TheShotRegistry->preAlloc(min(1024, 4096 * req_per_client));  // XXX: guess!
	// 1000 "middle" xactions for client; on-demand for server
	TheXactLog.configure(
		client_count ? (client_count*req_per_client / 2) : -1, 1000);

	if (client_count*req_per_client > 0) {
		const int trace_count = min(client_count*req_per_client, 10*1024);
		TheRespTimeTrace = new Trace("Response time", "usec", trace_count);
		TheRespSizeTrace = new Trace("Response size", "byte", trace_count);
	}

	TheShotRegistry->startTime(TheClock.now());

	// create and start some servers
	Array<Server*> servers;
	for (int i = 0; i < server_count; ++i) {
		Server *s = new Server();
		s->configure(scfg.clone());
		s->port(ports[i]);
		servers.append(s);
		TheKernel.start(s);
	}

	// create and start some clients
	Array<Client*> clients;
	for (int i = 0; i < client_count; ++i) {
		Client *c = new Client();
		c->configure(ccfg.clone());
		clients.append(c);
		TheKernel.start(clients[i]);
	}

	loop();

	UpdateClocks(false);
	Journal << "all jobs completed in " << (TheClock-startTime) << ende;
	Report << "execution time: " << (TheClock-startTime) << endl;

	// collect final reports
	MeterShot shot = ThePhaseMgr.mshot();
	TheErrHist.report(Report << endl);
	Report << endl;
	reportRespTimeAndSize();
	report("server", servers);
	report("client", clients);
	report(-1, shot.subm(), shot.compl(), shot.err());

	saveResults();

	// cleanup
	{
		for (int i = 0; i < clients.count(); ++i)
			delete clients[i];
		for (int i = 0; i < servers.count(); ++i) {
			TheKernel.stop(servers[i]);
			delete servers[i];
		}
		scfg.destroy();
		ccfg.destroy();

		delete TheRespTimeTrace;
		delete TheRespSizeTrace;
	}

	Journal << ThePrgName << " done." << ende;

	return 0;
}
